irc_chan = {}
irc_chan.name = ""
irc_chan.topic = ""
irc_chan.num_users = 0

function new_channel(name)
	local t = {}
	setmetatable(t, {__index = irc_chan})

	t.user_list = {}
	t.modes = {}
	t.name = name

	t.modes["+o"] = true
	t.modes["-o"] = true
	t.modes["+b"] = true
	t.modes["q"] = true
	t.modes["@"] = true

	return t
end

function irc_chan:joinattempt(id)
	return true
end

function irc_chan:user_say(id,txt)
	if self.user_list[id] then
		if string.sub(self.user_list[id].mode, 1, 2) == "+b" then
			return false, "You're banned from this channel"
		end
		return true
	end
	return false, "You're not in this channel"
end

function irc_chan:messageto(id,txt)
	if irc_client[id] and self.user_list[id] then
		irc_server.raw_message(id, ":"..self.name.."!"..self.name.."@"..irc_server.hostname.." PRIVMSG "..irc_client[id].nick.." :"..txt)
	end
end

function irc_chan:message(txt)
	for id, user in pairs(self.user_list) do
		if irc_client[id] then
			irc_server.raw_message(id, ":"..self.name.."!"..self.name.."@"..irc_server.hostname.." PRIVMSG "..irc_client[id].nick.." :"..txt)
		end
	end
end

function irc_chan:mode_exists(m)
	if not m then return nil end
	return self.modes[string.sub(m, 1, 2)] ~= nil
end

function irc_chan:opmode(mode)
	if not mode then return false end
	if string.sub(mode, 1, 1) == "q" then
		-- Owner
		return true
	elseif string.sub(mode, 1, 2) == "+o" then
		-- Operator
		return true
	elseif string.sub(mode, 1, 1) == "@" then
		-- Operator
		return true
	end
end

function irc_chan:owner(id)
	if irc_client[id] then
		if irc_client[id].op then
			return true
		end

		if self.user_list[id] then
			local mode = self.user_list[id].mode
			return string.sub(mode, 1, 1) == "@"
		end
	end
end

function irc_chan:op(id)
	if irc_client[id] then
		if irc_client[id].op then
			return true
		end

		if self.user_list[id] then
			local mode = self.user_list[id].mode
			return self:opmode(mode)
		end
	end
end

function irc_chan:kick(id,reason)
	if self.user_list[id] then
		irc_server.client_cmd(id, "PART", self.name, "QUIT", reason or "kicked")
		self:remove(id)
	end
end

function irc_chan:remove(id)
	if self.user_list[id] then
		self.num_users = self.num_users - 1
		self.user_list[id] = nil

		if irc_client[id] then
			if irc_client[id].channels[self.name] then
				irc_client[id].channels[self.name] = nil
			end
		end

		if self.num_users == 0 then
			irc_server.channels[self.name] = nil
		end
	end
end

function irc_chan:add(id,mode)
	if irc_client[id] then
		if not irc_client[id].channels[self.name] then
			irc_client[id].channels[self.name] = self
		end

		self.user_list[id] = {
			mode = mode or ""
		}
		self.num_users = self.num_users + 1
	end
end

function irc_chan:close()
	for id, _ in pairs(self.user_list) do
		self:kick(id, "Closed")
	end
end

function irc_chan:setmode(id,mode)
	local mode_user = self.user_list[id]
	if mode_user then
		mode_user.mode = mode
		irc_server.send_user_message(id, self.name, "MODE "..self.name.." "..mode.." "..irc_client[id].nick)
	end
end
