-- TCP Client
irc_cl = {}

irc_cl.state = "UNREGISTERED"
irc_cl.nick = ""
irc_cl.user = ""
irc_cl.host = ""
irc_cl.host_name = ""
irc_cl.real_name = ""
irc_cl.port = 0

irc_cl.ping_requested = false
irc_cl.last_ping_req = 0
irc_cl.last_ping = 0
irc_cl.ping = 0

function new_client()
	local t = {}
	t.channels = {}

	setmetatable(t, {__index = irc_cl})
	return t
end

function irc_cl:ip()
	if self.socket then
		local ip = self.socket:getpeername()
		return ip or "0.0.0.0"
	end
end

function irc_cl:port()
	if self.socket then
		local ip, port = self.socket:getpeername()
		if ip and port then
			return port
		end
		return 0
	end
end

function irc_cl:close()
	if self.socket then
		self.socket:close()
	end
end

function irc_cl:hostname()
	if self.host_name:len() > 0 then
		return self.host_name
	end

	local host = socket.dns.tohostname(self:ip())
	if host then
		if host == "*" then
			if self:ip() == "127.0.0.1" then
				self.host_name = "*"
				return host
			end
			return nil
		else
			self.host_name = host
			return host
		end
	end
end

function irc_cl:addr()
	if self.user:len() > 0 and self.host_name:len() > 0 then
		local addr = self.user.."@"..self.host_name
		if not irc_addr_id[addr] then
			irc_addr_id[addr] = self.id
		end
		return addr
	end
end

function irc_cl:destroy()
end

function irc_cl:send(text)
	if self.socket then
		self.socket:send(text)
	end
end

function irc_cl:getline()
	if self.socket then
		return self.socket:receive()
	end
	return nil, "no socket"
end

function irc_cl:process()
	if self.state == "UNREGISTERED" then
		if string.len(self.nick) > 0 and string.len(self.user) > 0 then
			if self.host_name:len() > 0 then
				self.state = "REGISTERED"
				irc_server.send_message(self.id, "Welcome to "..irc_server.name..", "..self.nick.."!"..self.user.."@"..self.host, 1)
				irc_server.send_message(self.id, "Your host is "..self.host_name..", running version "..irc_server.version, 2)
				irc_server.send_message(self.id, "This server was created...", 3)
				irc_server.send_message(self.id, irc_server.hostname.." "..irc_server.version.." aAbBcCdDeEfFGhHiIjkKlLmMnNopPQrRsStUvVwWxXyYzZ0123459*@ bcdefFhiIklmnoPqstv", 4)
				irc_server.client_cmd(self.id, "MOTD", "")
			elseif not self.lookupmessage then
				irc_server.raw_message(self.id, "NOTICE AUTH :*** Looking up your hostname...")
				if self:hostname() then
					irc_server.raw_message(self.id, "NOTICE AUTH :*** Found your hostname")
				else
					irc_server.raw_message(self.id, "NOTICE AUTH :*** Couldn't find your hostname, using your IP instead")
					self.host_name = self:ip()
				end

				self:addr()
				self.lookupmessage = true
			end
		end
	end

	local line, er = self:getline()
	if er == "closed" then
		return irc_server.client_cmd(self.id, "QUIT", er.." from client")
	end
	if not line or line:len() == 0 then return nil end

	local line = line:match("%s*(.+)%s*")
	if string.len(self.nick) > 0 then
		irc_server.debug(self.state..": "..self.nick.." -> '"..line.."'")
	else
		irc_server.debug(self.state..": "..self:ip()..":"..self:port().." -> '"..line.."'")
	end

	local command, args = line:match("^%s*([^ ]+) *(.*)%s*$")

	if command and args then
		command = command:upper()
		irc_server.client_cmd(self.id, command, args)
	end
end

-- UDP Client
irc_cl_udp = new_client()
irc_cl_udp.udp = true

function irc_cl_udp:ip()
	return self._ip or "0.0.0.0"
end

function irc_cl_udp:port()
	return self._port or "0.0.0.0"
end

function irc_cl_udp:getline()
	if self._datagram:len() > 0 then
		local f = self._datagram:find(string.char(10)) or self._datagram:len()
		if f then
			local data = self._datagram:sub(1, f-1)
			self._datagram = self._datagram:sub(f+1)
			return data
		end
	end
	return nil, "empty"
end

function irc_cl_udp:send(text)
	if self._ip and self._port then
		self._mastersock:sendto(text, self._ip, self._port)
	end
end

function irc_cl_udp:close()
	if self._ip and self._port then
		irc_udp_adr_id[self._ip..":"..self._port] = nil
	end
end

function new_client_udp()
	local t = {}
	t.channels = {}

	setmetatable(t, {__index = irc_cl_udp})
	return t
end

-- IRC BOT
irc_bot = new_client()
irc_bot.state = "REGISTERED"
irc_bot.user = "Bot"
irc_bot.real_name = "Bot"
irc_bot.host = "localhost"
irc_bot.host_name = "localhost"
irc_bot.bot = true

function new_bot()
	local t = {}
	t.channels = {}

	setmetatable(t, {__index = irc_bot})
	return t
end

function irc_bot:ip()
	return "127.0.0.1"
end

function irc_bot:port()
	return 0
end

function irc_bot:close()
end

function irc_bot:send(txt)
	local nick, user, host, channel, message = txt:match("(%a+)!(%a+)@(.+) PRIVMSG (.+) :(.+)")
	if nick and user and host and channel and message then
		self:receivemessage(user_id(nick), channel, message:sub(1, message:len()-2))
	end
end

function irc_bot:receivemessage(id,channel,txt)
end

function irc_bot:say(txt,channel)
	if not txt or not channel then return false end
	if txt:len() <= 0 or channel:len() <= 0 then return false end
	irc_server.send_user_message(self.id, channel, "PRIVMSG "..channel.." :"..txt, true)
	return true
end

function irc_bot:joinchannel(channel)
	if not channel then return false end
	if channel:len() <= 0 then return false end
	irc_server.client_cmd(self.id, "JOIN", channel)
	return true
end

function irc_bot:partchannel(channel)
	if not channel then return false end
	if channel:len() <= 0 then return false end
	irc_server.client_cmd(self.id, "PART", channel)
	return true
end

function irc_bot:process()
end
