require("socket")

irc_server = {}
irc_dir = ""

irc_server.name = "Starkkz's IRC"
irc_server.hostname = "localhost"
irc_server.version = "0.0.0.1"

irc_server.motd = ""

irc_server.tcp_ports = {6667,6668}
irc_server.udp_ports = {6669}

irc_server.tcp_socket = {}
irc_server.udp_socket = {}

irc_server.channels = {}

irc_server.command = {}

irc_client = {}
irc_nick_id = {}
irc_addr_id = {}
irc_udp_adr_id = {}

function irc_server.read_motd()
	local f = io.open("motd.txt", "r")
	if f then
		for line in f:lines() do
			irc_server.motd = irc_server.motd.."\n"..line
		end
		f:close()
	else
		irc_server.debug("Failed to load MOTD")
	end
end; irc_server.read_motd()

function irc_server.find_host_id(user,host)
	for userhost, id in pairs(irc_addr_id) do
		local _user, _host = userhost:match("(.+)@(.+)")
		if user == "*" or user == _user then
			if host == _host then
				return id
			end
		end
	end
end

function irc_tab_nick(nick)
	local id = irc_nick_id[nick]
	if id then return irc_client[nick] end
end

function user_nick(id)
	if irc_client[id] then return irc_client[id].nick end
end

function user_id(nick)
	return irc_nick_id[nick]
end

irc_code = {}
function irc_server.generate_code(length)
	local length = length or 8
	local code = ""
	local scale = {{97, 122},{65,90},{48,57}}

	for i = 1, length do
		local x = math.random(1, #scale)
		local y = math.random(scale[x][1], scale[x][2])
		code = code .. string.char(y)
	end

	if irc_code[code] then
		return irc_server.generate_code(length)
	else
		irc_code[code] = true
		return code
	end
end

function irc_server.load_cmd(cmd,file,unregistered)
	local file = io.open(file)

	if file then
		local code = file:read("*a"); code = "local id, args, arg = ...\n"..code
		local f, e = loadstring(code)
		file:close()

		if f then
			irc_server.command[cmd] = {f,unregistered}
		elseif e then
			print("LUA ERROR ("..cmd.."): "..e)
		end
	end
end

function irc_server.delete_code(code)
	if code then
		irc_code[code] = nil
	end
end

function irc_server.set_nick(id,nick)
	if irc_client[id] then
		if irc_client[id].nick:len() > 0 then
			irc_nick_id[irc_client[id].nick] = nil
		end

		irc_nick_id[nick] = id
		irc_client[id].nick = nick
	end
end

function init_tcp_port(port)
	local s = socket.tcp()
	local r, e = s:bind("*", port)
	if not e then
		s:settimeout(0)
		s:listen()
		irc_server.debug("Initialized TCP server with port "..port)
		return s
	else
		irc_server.debug("Failed to start TCP server ("..port..") "..e)
	end
end

function init_udp_port(port)
	local s = socket.udp()
	local r, e = s:setsockname("*", port)
	if r and r == 1 then
		s:settimeout(0)
		irc_server.debug("Initialized UDP server with port "..port)
		return s
	else
		irc_server.debug("Failed to start UDP server ("..port..") "..e)
	end
end

function tcp_server(port)
	local i = #irc_server.tcp_socket + 1
	irc_server.tcp_socket[i] = init_tcp_port(port)
end

function udp_server(port)
	local i = #irc_server.udp_socket + 1
	irc_server.udp_socket[i] = init_udp_port(port)
end

function irc_server.init_ports()
	irc_server.debug("Initializing ports")
	for i = 1, #irc_server.tcp_ports do
		irc_server.tcp_socket[i] = init_tcp_port(irc_server.tcp_ports[i])
	end

	for i = 1, #irc_server.udp_ports do
		irc_server.udp_socket[i] = init_udp_port(irc_server.udp_ports[i])
	end
end

function irc_server.raw_message(id, txt)
	if irc_client[id] then
		irc_client[id]:send(txt.."\r\n")
	end
end

function irc_server.client_cmd(id,cmd,args,...)
	local t = irc_server.command[cmd]
	if t then
		local f = t[1]
		local unregistered = t[2]
		if f and type(f) == "function" then
			if unregistered or (not unregistered and irc_client[id].state == "REGISTERED") then
				f(id, args, arg)
			end
		end
	end
end

function irc_server.send_message(id, txt, code)
	if irc_client[id] then
		if not code then code = "000" end
		if string.len(code) == 2 then
			code = "0"..code
		elseif string.len(code) == 1 then
			code = "00"..code
		end
		irc_server.raw_message(id, ":"..irc_server.hostname.." "..code.." "..irc_client[id].nick.." :"..txt)
	end
end

function irc_server.send_user_message(id, target, text, noSelf)
	local address = irc_client[id]:addr()
	if not address then return nil end

	local user_str = ":"..irc_client[id].nick.."!"..address
	if string.sub(target, 1, 1) == "#" then
		local channel_info = irc_server.channels[target]
		if channel_info then
			for user, _ in pairs(channel_info.user_list) do
				if noSelf and user == id then
					-- Do not send
				else
					irc_server.raw_message(user, user_str.." "..text)
				end
			end
		end
	elseif type(target) == "string" then
		local u = irc_nick_id[target]
		if u then
			irc_server.raw_message(u, user_str.." "..text)
		end
	elseif type(target) == "number" then
		irc_server.raw_message(target, user_str.." "..text)
	end
end

function irc_server.generate_nick()
	local i = 1
	while irc_nick_id["User_"..i] do
		i = i + 1
	end
	return "User_"..i
end

function irc_server.accept_cl(s)
	s:settimeout(0, "b")
	s:settimeout(0, "t")

	local id = #irc_client + 1
	irc_client[id] = new_client()
	irc_client[id].id = id
	irc_client[id].socket = s

	irc_server.debug("New TCP client "..irc_client[id]:ip()..":"..irc_client[id]:port())
end

function irc_server.accept_tcp()
	for i, v in pairs(irc_server.tcp_socket) do
		if irc_server.tcp_socket[i] then
			local s = irc_server.tcp_socket[i]:accept()
			while s do
				irc_server.accept_cl(s)
				s = irc_server.tcp_socket[i]:accept()
			end
		end
	end
end

function irc_server.accept_udp()
	for i, v in pairs(irc_server.udp_socket) do
		local line, ip, port = irc_server.udp_socket[i]:receivefrom()
		if line and ip and port then
			local id = irc_udp_adr_id[ip..":"..port]
			if not id or not irc_client[id] then
				-- Received message from unregistered
				id = #irc_client + 1
				irc_client[id] = new_client_udp()
				irc_client[id].id = id
				irc_client[id]._ip = ip
				irc_client[id]._port = port
				irc_client[id]._mastersock = irc_server.udp_socket[i]
				irc_client[id]._datagram = ""
				irc_udp_adr_id[ip..":"..port] = id

				irc_server.debug("New UDP client "..irc_client[id]:ip()..":"..irc_client[id]:port())
			end

			if irc_client[id] then
				irc_client[id]._datagram = irc_client[id]._datagram .. line
			end
		end
	end
end

function irc_server.update()
	local t = socket.gettime()

	for id, client in pairs(irc_client) do
		client:process()

		if not client.bot and client.udp then
			if t - client.last_ping >= 20000 and client.ping_requested then
				irc_server.client_cmd(id, "QUIT", "timeout")
			elseif t - client.last_ping >= 6000 then
				if not client.ping_code then
					client.ping_code = irc_server.generate_code()
				end

				if t - client.last_ping_req >= 500 then
					client.ping_requested = true
					client.last_ping_req = t
					irc_server.raw_message(id, "PING "..client.ping_code)
				end
			end
		end
	end

	irc_server.accept_tcp()
	irc_server.accept_udp()
end

function irc_server.debug(txt)
	print("[IRC] "..txt)
end

function irc_server.add_bot(nick,channel)
	if not irc_nick_id[nick] then
		local id = #irc_client + 1
		irc_client[id] = new_bot()
		irc_client[id].id = id
		irc_server.set_nick(id, nick)
		irc_client[id].channel_name = channel
		irc_server.client_cmd(id, "JOIN", channel)
		irc_server.debug("New bot "..nick.." in channel "..channel)

		irc_nick_id[nick] = id
		return id
	else
		irc_server.debug("Failed to add bot, this name is already in use")
	end
	return -1
end

dofile(irc_dir.."command_config.lua")
dofile(irc_dir.."client.lua")
dofile(irc_dir.."channel.lua")

irc_server.init_ports()

--[[
enum Reply
    {
        RplNone             = 0,
        // Initial
        RplWelcome          = 001,                  // :Welcome to the Internet Relay Network <nickname>
        RplYourHost         = 002,                  // :Your host is <server>, running version <ver>
        RplCreated          = 003,                  // :This server was created <datetime>
        RplMyInfo           = 004,                  // <server> <ver> <usermode> <chanmode>
        RplMap              = 005,                  // :map
        RplEndOfMap         = 007,                  // :End of /MAP
        RplMotdStart        = 375,                  // :- server Message of the Day
        RplMotd             = 372,                  // :- <info>
        RplMotdAlt          = 377,                  // :- <info>                                                                        (some)
        RplMotdAlt2         = 378,                  // :- <info>                                                                        (some)
        RplMotdEnd          = 376,                  // :End of /MOTD command.
        RplUModeIs          = 221,                  // <mode>

        // IsOn/UserHost
        RplUserHost         = 302,                  // :userhosts
        RplIsOn             = 303,                  // :nicknames

        // Away
        RplAway             = 301,                  // <nick> :away
        RplUnAway           = 305,                  // :You are no longer marked as being away
        RplNowAway          = 306,                  // :You have been marked as being away

        // WHOIS/WHOWAS
        RplWhoisHelper      = 310,                  // <nick> :looks very helpful                                                       DALNET
        RplWhoIsUser        = 311,                  // <nick> <username> <address> * :<info>
        RplWhoIsServer      = 312,                  // <nick> <server> :<info>
        RplWhoIsOperator    = 313,                  // <nick> :is an IRC Operator
        RplWhoIsIdle        = 317,                  // <nick> <seconds> <signon> :<info>
        RplEndOfWhois       = 318,                  // <request> :End of /WHOIS list.
        RplWhoIsChannels    = 319,                  // <nick> :<channels>
        RplWhoWasUser       = 314,                  // <nick> <username> <address> * :<info>
        RplEndOfWhoWas      = 369,                  // <request> :End of WHOWAS
        RplWhoReply         = 352,                  // <channel> <username> <address> <server> <nick> <flags> :<hops> <info>
        RplEndOfWho         = 315,                  // <request> :End of /WHO list.
        RplUserIPs          = 307,                  // :userips                                                                         UNDERNET
        RplUserIP           = 340,                  // <nick> :<nickname>=+<user>@<IP.address>                                          UNDERNET

        // List
        RplListStart        = 321,                  // Channel :Users Name
        RplList             = 322,                  // <channel> <users> :<topic>
        RplListEnd          = 323,                  // :End of /LIST
        RplLinks            = 364,                  // <server> <hub> :<hops> <info>
        RplEndOfLinks       = 365,                  // <mask> :End of /LINKS list.

        // Post-Channel Join
        RplUniqOpIs         = 325,
        RplChannelModeIs    = 324,                  // <channel> <mode>
        RplChannelUrl       = 328,                  // <channel> :url                                                                   DALNET
        RplChannelCreated   = 329,                  // <channel> <time>
        RplNoTopic          = 331,                  // <channel> :No topic is set.
        RplTopic            = 332,                  // <channel> :<topic>
        RplTopicSetBy       = 333,                  // <channel> <nickname> <time>
        RplNamReply         = 353,                  // = <channel> :<names>
        RplEndOfNames       = 366,                  // <channel> :End of /NAMES list.

        // Invitational
        RplInviting         = 341,                  // <nick> <channel>
        RplSummoning        = 342,

        // Channel Lists
        RplInviteList       = 346,                  // <channel> <invite> <nick> <time>                                                 IRCNET
        RplEndOfInviteList  = 357,                  // <channel> :End of Channel Invite List                                            IRCNET
        RplExceptList       = 348,                  // <channel> <exception> <nick> <time>                                              IRCNET
        RplEndOfExceptList  = 349,                  // <channel> :End of Channel Exception List                                         IRCNET
        RplBanList          = 367,                  // <channel> <ban> <nick> <time>
        RplEndOfBanList     = 368,                  // <channel> :End of Channel Ban List


        // server/misc
        RplVersion          = 351,                  // <version>.<debug> <server> :<info>
        RplInfo             = 371,                  // :<info>
        RplEndOfInfo        = 374,                  // :End of /INFO list.
        RplYoureOper        = 381,                  // :You are now an IRC Operator
        RplRehashing        = 382,                  // <file> :Rehashing
        RplYoureService     = 383,
        RplTime             = 391,                  // <server> :<time>
        RplUsersStart       = 392,
        RplUsers            = 393,
        RplEndOfUsers       = 394,
        RplNoUsers          = 395,
        RplServList         = 234,
        RplServListEnd      = 235,
        RplAdminMe          = 256,                  // :Administrative info about server
        RplAdminLoc1        = 257,                  // :<info>
        RplAdminLoc2        = 258,                  // :<info>
        RplAdminEMail       = 259,                  // :<info>
        RplTryAgain         = 263,                  // :Server load is temporarily too heavy. Please wait a while and try again.

        // tracing
        RplTraceLink        = 200,
        RplTraceConnecting  = 201,
        RplTraceHandshake   = 202,
        RplTraceUnknown     = 203,
        RplTraceOperator    = 204,
        RplTraceUser        = 205,
        RplTraceServer      = 206,
        RplTraceService     = 207,
        RplTraceNewType     = 208,
        RplTraceClass       = 209,
        RplTraceReconnect   = 210,
        RplTraceLog         = 261,
        RplTraceEnd         = 262,

        // stats
        RplStatsLinkInfo    = 211,                  // <connection> <sendq> <sentmsg> <sentbyte> <recdmsg> <recdbyte> :<open>
        RplStatsCommands    = 212,                  // <command> <uses> <bytes>
        RplStatsCLine       = 213,                  // C <address> * <server> <port> <class>
        RplStatsNLine       = 214,                  // N <address> * <server> <port> <class>
        RplStatsILine       = 215,                  // I <ipmask> * <hostmask> <port> <class>
        RplStatsKLine       = 216,                  // k <address> * <username> <details>
        RplStatsPLine       = 217,                  // P <port> <??> <??>
        RplStatsQLine       = 222,                  // <mask> :<comment>
        RplStatsELine       = 223,                  // E <hostmask> * <username> <??> <??>
        RplStatsDLine       = 224,                  // D <ipmask> * <username> <??> <??>
        RplStatsLLine       = 241,                  // L <address> * <server> <??> <??>
        RplStatsuLine       = 242,                  // :Server Up <num> days, <time>
        RplStatsoLine       = 243,                  // o <mask> <password> <user> <??> <class>
        RplStatsHLine       = 244,                  // H <address> * <server> <??> <??>
        RplStatsGLine       = 247,                  // G <address> <timestamp> :<reason>
        RplStatsULine       = 248,                  // U <host> * <??> <??> <??>
        RplStatsZLine       = 249,                  // :info
        RplStatsYLine       = 218,                  // Y <class> <ping> <freq> <maxconnect> <sendq>
        RplEndOfStats       = 219,                  // <char> :End of /STATS report
        RplStatsUptime      = 242,

        // GLINE
        RplGLineList        = 280,                  // <address> <timestamp> <reason>                                                   UNDERNET
        RplEndOfGLineList   = 281,                  // :End of G-line List                                                              UNDERNET

        // Silence
        RplSilenceList      = 271,                  // <nick> <mask>                                                                    UNDERNET/DALNET
        RplEndOfSilenceList = 272,                  // <nick> :End of Silence List                                                      UNDERNET/DALNET

        // LUser
        RplLUserClient      = 251,                  // :There are <user> users and <invis> invisible on <serv> servers
        RplLUserOp          = 252,                  // <num> :operator(s) online
        RplLUserUnknown     = 253,                  // <num> :unknown connection(s)
        RplLUserChannels    = 254,                  // <num> :channels formed
        RplLUserMe          = 255,                  // :I have <user> clients and <serv> servers
        RplLUserLocalUser   = 265,                  // :Current local users: <curr> Max: <max>
        RplLUserGlobalUser  = 266,                  // :Current global users: <curr> Max: <max>


        // Errors
        ErrNoSuchNick       = 401,                  // <nickname> :No such nick
        ErrNoSuchServer     = 402,                  // <server> :No such server
        ErrNoSuchChannel    = 403,                  // <channel> :No such channel
        ErrCannotSendToChan = 404,                  // <channel> :Cannot send to channel
        ErrTooManyChannels  = 405,                  // <channel> :You have joined too many channels
        ErrWasNoSuchNick    = 406,                  // <nickname> :There was no such nickname
        ErrTooManyTargets   = 407,                  // <target> :Duplicate recipients. No message delivered
        ErrNoColors         = 408,                  // <nickname> #<channel> :You cannot use colors on this channel. Not sent: <text>   DALNET
        ErrNoOrigin         = 409,                  // :No origin specified
        ErrNoRecipient      = 411,                  // :No recipient given (<command>)
        ErrNoTextToSend     = 412,                  // :No text to send
        ErrNoTopLevel       = 413,                  // <mask> :No toplevel domain specified
        ErrWildTopLevel     = 414,                  // <mask> :Wildcard in toplevel Domain
        ErrBadMask          = 415,
        ErrTooMuchInfo      = 416,                  // <command> :Too many lines in the output, restrict your query                     UNDERNET
        ErrUnknownCommand   = 421,                  // <command> :Unknown command
        ErrNoMotd           = 422,                  // :MOTD File is missing
        ErrNoAdminInfo      = 423,                  // <server> :No administrative info available
        ErrFileError        = 424,
        ErrNoNicknameGiven  = 431,                  // :No nickname given
        ErrErroneusNickname = 432,                  // <nickname> :Erroneus Nickname
        ErrNickNameInUse    = 433,                  // <nickname> :Nickname is already in use.
        ErrNickCollision    = 436,                  // <nickname> :Nickname collision KILL
        ErrUnAvailResource  = 437,                  // <channel> :Cannot change nickname while banned on channel
        ErrNickTooFast      = 438,                  // <nick> :Nick change too fast. Please wait <sec> seconds.                         (most)
        ErrTargetTooFast    = 439,                  // <target> :Target change too fast. Please wait <sec> seconds.                     DALNET/UNDERNET
        ErrUserNotInChannel = 441,                  // <nickname> <channel> :They aren't on that channel
        ErrNotOnChannel     = 442,                  // <channel> :You're not on that channel
        ErrUserOnChannel    = 443,                  // <nickname> <channel> :is already on channel
        ErrNoLogin          = 444,
        ErrSummonDisabled   = 445,                  // :SUMMON has been disabled
        ErrUsersDisabled    = 446,                  // :USERS has been disabled
        ErrNotRegistered    = 451,                  // <command> :Register first.
        ErrNeedMoreParams   = 461,                  // <command> :Not enough parameters
        ErrAlreadyRegistered= 462,                  // :You may not reregister
        ErrNoPermForHost    = 463,
        ErrPasswdMistmatch  = 464,
        ErrYoureBannedCreep = 465,
        ErrYouWillBeBanned  = 466,
        ErrKeySet           = 467,                  // <channel> :Channel key already set
        ErrServerCanChange  = 468,                  // <channel> :Only servers can change that mode                                     DALNET
        ErrChannelIsFull    = 471,                  // <channel> :Cannot join channel (+l)
        ErrUnknownMode      = 472,                  // <char> :is unknown mode char to me
        ErrInviteOnlyChan   = 473,                  // <channel> :Cannot join channel (+i)
        ErrBannedFromChan   = 474,                  // <channel> :Cannot join channel (+b)
        ErrBadChannelKey    = 475,                  // <channel> :Cannot join channel (+k)
        ErrBadChanMask      = 476,
        ErrNickNotRegistered= 477,                  // <channel> :You need a registered nick to join that channel.                      DALNET
        ErrBanListFull      = 478,                  // <channel> <ban> :Channel ban/ignore list is full
        ErrNoPrivileges     = 481,                  // :Permission Denied- You're not an IRC operator
        ErrChanOPrivsNeeded = 482,                  // <channel> :You're not channel operator
        ErrCantKillServer   = 483,                  // :You cant kill a server!
        ErrRestricted       = 484,                  // <nick> <channel> :Cannot kill, kick or deop channel service                      UNDERNET
        ErrUniqOPrivsNeeded = 485,                  // <channel> :Cannot join channel (reason)
        ErrNoOperHost       = 491,                  // :No O-lines for your host
        ErrUModeUnknownFlag = 501,                  // :Unknown MODE flag
        ErrUsersDontMatch   = 502,                  // :Cant change mode for other users
        ErrSilenceListFull  = 511                   // <mask> :Your silence list is full                                                UNDERNET/DALNET

    };  // eo enum Reply
]]
